#include "material.h"
#include <qmessagebox.h>

Material::Material()
{
  setSpecularCoefficient(0.0);
  setTextureScale(1.0);
  refractiveIndex_=100;
  reflectiveColor_=refractiveColor_=Color(0,0,0);
}

void Material::initFromDOMElement(const QDomElement& e)
{
  if (e.tagName() != "Material")
    {
      QMessageBox::critical(NULL, "Material init error", "Material::initFromDOMElement, bad DOM tagName.\nExpecting 'Material', got "+e.tagName());
      return;
    }

  if(e.hasAttribute("refractiveIndex"))
	  this->setRefractiveIndex(e.attribute("refractiveIndex", "0").toFloat());
 /* else
	QMessageBox::warning(NULL, "Material warning", "No Material refractive coefficient provided. Using 0.0 instead");
*/
  QDomNode n = e.firstChild();

  while (!n.isNull()) {

      QDomElement e = n.toElement();
      if (!e.isNull()) {

		  if (e.tagName() == "DiffuseColor")
			  diffuseColor_.initFromDOMElement(e);

		  else if (e.tagName() == "SpecularColor") {
			  specularColor_.initFromDOMElement(e);
			  if (!e.hasAttribute("exponent"))
				  QMessageBox::warning(NULL, "Material error", "No Material specular coefficient provided. Using 0.0 instead");
			  setSpecularCoefficient(e.attribute("exponent", "0.0").toFloat());
		  }

		  else if (e.tagName() == "ReflectiveColor")
			  reflectiveColor_.initFromDOMElement(e);

	      else if(e.tagName() == "RefractiveColor") {
			  refractiveColor_.initFromDOMElement(e); 

			  /*if (!e.hasAttribute("refractiveIndex"))
				  QMessageBox::warning(NULL, "Material error", "No Material specular coefficient provided. Using 0.0 instead");
			  setRefractiveIndex(e.attribute("refractiveIndex", "0.0").toFloat()); */

			  if (e.hasAttribute("refractiveIndex"))
				  QMessageBox::warning(NULL, "Material error", "Refrative index must be set as material attribute");
				  

		  }
		  
		  else if (e.tagName() == "Texture") {
			  setTextureScale(e.attribute("scale", "1.0").toFloat());
			  if (e.hasAttribute("scaleU")) 
				  setTextureScaleU(e.attribute("scaleU", "1.0").toFloat());
			  if (e.hasAttribute("scaleV"))
				  setTextureScaleV(e.attribute("scaleV", "1.0").toFloat());
			  loadTextureFromFile(this->texture_, e.attribute("file"));

			  if (e.attribute("mode") == "MODULATE")
				  setTextureMode(MODULATE);
			  if (e.attribute("mode") == "BLEND")
				  setTextureMode(BLEND);
			  if (e.attribute("mode") == "REPLACE")
				  setTextureMode(REPLACE);
		  }

		  // "Extension: Perturbation de normales"
		  else if (e.tagName() == "BumpMapTexture") {
			  setBumpMapScale(e.attribute("scale", "1.0").toFloat());
			  if (e.hasAttribute("scaleU")) 
				  setBumpMapScaleU(e.attribute("scaleU", "1.0").toFloat());
			  if (e.hasAttribute("scaleV"))
				  setBumpMapScaleV(e.attribute("scaleV", "1.0").toFloat());
			  if (e.hasAttribute("height"))
				  setBumpMapHeight(e.attribute("height", "1.0").toFloat());
			  loadTextureFromFile(this->bumpMapTexture_, e.attribute("file"));
		  }
		  else if (e.tagName() == "SpecularMapTexture") {
			  setSpecularMapScale(e.attribute("scale", "1.0").toFloat());
			  if (e.hasAttribute("scaleU")) 
				  setSpecularMapScaleU(e.attribute("scaleU", "1.0").toFloat());
			  if (e.hasAttribute("scaleV"))
				  setSpecularMapScaleV(e.attribute("scaleV", "1.0").toFloat());
			  loadTextureFromFile(this->specularMapTexture_, e.attribute("file"));
		  }


		  else
			  QMessageBox::warning(NULL, "Material child error", "Unsupported Material child : "+e.tagName()); 
	  }
      else
		  QMessageBox::warning(NULL, "Material XML error", "Error while parsing Material XML document");

      n = n.nextSibling();
    }
}

void Material::loadTextureFromFile(QImage &qi, const QString& fileName)
{
  if (fileName.isNull())
    QMessageBox::warning(NULL, "Material texture error", "Material error : no texture file name provided");
  else
    if (!qi.load(fileName))
      QMessageBox::warning(NULL, "Material texture error", "Unable to load Material texture from "+fileName);
}

Color Material::diffuseColor(float u, float v) const
{
  if (texture().isNull())
    return diffuseColor();

  // Take texture scaling into account
  u /= textureScaleU();
  v /= textureScaleV();

  u=fabs(u - floor(u));
  v=fabs(v - floor(v));
  
  Color colPix;

  // Compute pixel coordinates
  /*const int i = int(fabs(u - floor(u)) * texture().width());
  const int j = int(fabs(v - floor(v)) * texture().height());
  colPix=Color(texture().pixel(i,j));*/

  float alpha=1.0f;

  //Ext float textures
  colPix=imageInterpValue(texture(), fabs(u)* texture().width(), fabs(v)* texture().height(), &alpha);

  // Compute color
  switch(textureMode())
    {
    case MODULATE : 
		//return Color(texture().pixel(i,j)) * diffuseColor();
		//Ext float textures
		return colPix * diffuseColor();
		
    case BLEND    :
      {
		//const float alpha = colPix.intensity();
		return alpha*colPix + (1.0f-alpha)*diffuseColor();
      }
    case REPLACE  : return colPix;
    }
  return Color();
}


// "Extension: Perturbation de normales"
/*qglviewer::Vec Material::bumpMapColor(float u, float v) const
{

  if (bumpMapTexture().isNull())
    return qglviewer::Vec(-1,-1,-1);


  // Take texture scaling into account
  u /= bumpMapScaleU();
  v /= bumpMapScaleV();

  // Compute pixel coordinates
  float i = fabs(u) * bumpMapTexture().width();
  float j = fabs(v) * bumpMapTexture().height();

  

  // Compute color
  qglviewer::Vec normal;


// 1st method: from bump map
  float iminus = i-1;
  float iplus = i+1;
  float jminus = j-1;
  float jplus = j+1;

  if (iminus < 0.0f)
	  iminus = 0.0f;
  if (jminus < 0.0f)
	  jminus = 0.0f;
  if (iplus >= bumpMapTexture().width())
	  iplus = bumpMapTexture().width()-1.0f;
  if (jplus >= bumpMapTexture().height() )
	  jplus = bumpMapTexture().height()-1.0f;

  
  Color bumpMapColPx1 = imageInterpValue(bumpMapTexture(),iminus, j);
  Color bumpMapColPy1 =imageInterpValue(bumpMapTexture(),i,jminus);
  Color bumpMapColPx = imageInterpValue(bumpMapTexture(),iplus,j);
  Color bumpMapColPy = imageInterpValue(bumpMapTexture(),i,jplus);

  float Dx = (bumpMapColPx.r - bumpMapColPx1.r)*5;
  float Dy = (bumpMapColPy.r - bumpMapColPy1.r)*5;
  float Dz2 = 1 - Dx*Dx - Dy*Dy; 
  if (Dz2 < 0)
	  Dz2 = 0;
  float Dz = sqrt(Dz2);

  normal.x = Dx;
  normal.y = -Dy;
  normal.z = Dz;


  return normal;
  
}*/


qglviewer::Vec Material::bumpMapVec(float u, float v) const
{

  if (bumpMapTexture().isNull())
    return qglviewer::Vec(0,0,0);


  // Take texture scaling into account
  u /= bumpMapScaleU();
  v /= bumpMapScaleV();

  
  u=fabs(u - floor(u));
  v=fabs(v - floor(v));

   // Compute pixel coordinates
  float i = fabs(u) * bumpMapTexture().width();
  float j = fabs(v) * bumpMapTexture().height();


  // Compute color

// 1st method: from bump map
  float iminus = i-1;
  float iplus = i+1;
  float jminus = j-1;
  float jplus = j+1;

  if (iminus < 0.0f)
	  iminus = 0.0f;
  if (jminus < 0.0f)
	  jminus = 0.0f;
  if (iplus >= bumpMapTexture().width())
	  iplus = bumpMapTexture().width()-1.0f;
  if (jplus >= bumpMapTexture().height() )
	  jplus = bumpMapTexture().height()-1.0f;

  
  Color bumpMapColPx1 = imageInterpValue(bumpMapTexture(),iminus, j);
  Color bumpMapColPy1 =imageInterpValue(bumpMapTexture(),i,jminus);
  Color bumpMapColPx = imageInterpValue(bumpMapTexture(),iplus,j);
  Color bumpMapColPy = imageInterpValue(bumpMapTexture(),i,jplus);


  qglviewer::Vec tangent = qglviewer::Vec(1, 0, (bumpMapColPx.r - bumpMapColPx1.r)*bumpMapHeight());
  qglviewer::Vec bitangent = qglviewer::Vec(0, 1, (bumpMapColPy.r - bumpMapColPy1.r)*bumpMapHeight());

    
  qglviewer::Vec normalTS=(tangent^bitangent).unit();
  normalTS.y=-normalTS.y;

  return normalTS;
  
}


// "Extension: Perturbation de normales"
float Material::specularMapColor(float u, float v) const
{

  if (specularMapTexture().isNull())
    return 1.0f;

  // Take texture scaling into account
  u /= specularMapScaleU();
  v /= specularMapScaleV();

  u=fabs(u - floor(u));
  v=fabs(v - floor(v));
  
  Color colPix;

  // Compute pixel coordinates
  /*const int i = int(fabs(u - floor(u)) * specularMapTexture().width());
  const int j = int(fabs(v - floor(v)) * specularMapTexture().height());
  colPix=Color(specularMapTexture().pixel(i,j));*/

  //Ext float textures
  colPix=imageInterpValue(specularMapTexture(), fabs(u)* specularMapTexture().width(), fabs(v)* specularMapTexture().height());

  return colPix.r;  
}

Color Material::imageInterpValue(const QImage &qi, float u, float v, float *alpha) const{
	u=abs(u);
	v=abs(v);

 int i = int(floor(u))%qi.width();
 int j = int(floor(v))%qi.height();

 int ip=(i+1)%qi.width();
 int jp=(j+1)%qi.height();

 Color v0=Color(qi.pixel(i, j));
 Color v1=Color(qi.pixel(ip, j));
 Color v2=Color(qi.pixel(i, jp));
 Color v3=Color(qi.pixel(ip, jp));

 float du=u-(float)i;
 float dv=v-(float)j;

 Color res= (1.0f-dv)*((1.0f-du)*v0+(du)*v1) + (dv)*((1.0f-du)*v2+(du)*v3);

 if(alpha){
	float a0=qAlpha(qi.pixel(i, j)) / 255.0f;
	float a1=qAlpha(qi.pixel(ip, j)) / 255.0f;
	float a2=qAlpha(qi.pixel(i, jp)) / 255.0f;
	float a3=qAlpha(qi.pixel(ip, jp)) / 255.0f;

	*alpha=(1.0f-dv)*((1.0f-du)*a0+(du)*a1) + (dv)*((1.0f-du)*a2+(du)*a3);

 }
 return res;
}